//Codec of file is UTF-8
//-----------------------------------------------------------------------
//      Решение дифференциальных уравнений в частных производных
//
// 
// Составлено по учебнику Алексеева, адаптировано под SciLab 5.5.2
//-----------------------------------------------------------------------


// "Ручная" реализация метода сеток решения дифференциальных уравнений в 
// частных производных. 

// Далее будет описана реализация метода сеток для решения параболических, 
// гиперболических и эллиптических уравнений.

// Решить параболическое уравнение, описывающее распределение температуры 
// в стержне длиной L, начальная температура стержня задается произвольной 
// функцией fi(x). Температуры концов стержня равны u(0,t)=U1=const, 
// u(L,t)=U2=const

// Подпрограмма решения задачи с помощью явной разностной схемы

//Правая часть дифференциального уравнения.
function y=f(x,t)
    y=sin(x*t)
endfunction

//Начальное условие
function y=fi(x)
    y=exp(0.15*x)
endfunction

//Условие на левой границе
function y=myu(t)
    y=1
endfunction

//Условие на правой границе
function y=nyu(x)
    y=2.117
endfunction

function [u,x,t]=parabol(N,K,L,T,a)
// Функция решения параболического уравнения методом сеток с
// помощью явной разностной схемы. N - количество участков,
// на которые разбивается интервал по х (0,L); K – количество
// участков, на которые разбивается интервал по t (0,T); a -
// параметр дифференциального уравнения теплопроводности,
// Функция возвращает матрицу решений u и вектора x, t

// Вычисляем шаг по х
h=L/N;
// Вычисляем шаг по t
delta=T/K;
// Формируем массив x и первый столбец матрицы решений U из начального 
// условия
for i=1:N+1
  x(i)=(i-1)*h;
  u(i,1)=fi(x(i));
end
// Формируем массив t, первую и последнюю строку матрицы решений U из 
// граничных условий
for j=1:K+1
  t(j)=(j-1)*delta;
  u(1,j)=myu(t(j));
  u(N+1,j)=nyu(t(j));
end
gam=a^2*delta/h^2;
// Формируем матрицу решений u с помощью явной разностной схемы
for j=1:K
    for i=2:N
        u(i,j+1)=gam*u(i-1,j)+(1-2*gam)*u(i,j)+gam*u(i+1,j)+....
            delta*f(x(i),t(j));
    end
end
endfunction

[U,X,T]=parabol(50,200,5,3,0.4);
surf(X,T,U');
title('PARABOLIC EQUATION (var.1)');
xlabel('X');
ylabel('T');

halt;
clear;

// При решении параболических уравнений с помощью явной разностной схемы 
// основной проблемой является устойчивость решения и правильный выбор 
// шага. Для решения этой проблемы были предложены неявные разностные
// схемы. Эти схемы абсолютно устойчивы, но алгоритм решения получаемого 
// сеточного уравнения несколько сложнее. Рассмотрим неявную разностную 
// схему для параболического уравнения.

//Правая часть дифференциального уравнения
function y=f(x,t)
    y=sin(x*t)
endfunction

//Начальное условие
function y=fi(x)
    y=exp(0.15*x)
endfunction

//Условие на левой границе
function y=myu(t)
    y=1
endfunction

//Условие на правой границе
function y=nyu(x)
    y=2.117
endfunction

function [u,x,t,r,k]=neyavn(N,K,L,T,a,eps)
// Функция решения параболического уравнения методом сеток с помощью 
// неявной разностной схемы. 
// N - количество участков, на которые разбивается интервал по х (0,L);
// K - количество участков, на которые разбивается интервал по t (0,T);
// a - параметр дифференциального уравнения теплопроводности, 
// eps - точность решения СЛАУ методом Зейделя.
// Функция neyavn возвращает:
// u - матрицу решений в узлах сетки, массив x, массив t,
// r - точность решения системы (12.14) методом Зейделя, 
// к - количество итераций при решении системы (7.14) методом Зейделя.

// Вычисляем шаг по х
h=L/N;
// Вычисляем шаг по t
delta=T/K;
// Формируем массив x и первый столбец матрицы решений U из начального 
// условия
for i=1:N+1
    x(i)=(i-1)*h;
    u(i,1)=fi(x(i));
end
// Формируем массив t, первую и последнюю строку матрицы решений U из 
// граничных условий
for j=1:K+1
    t(j)=(j-1)*delta;
    u(1,j)=myu(t(j));
    u(N+1,j)=nyu(t(j));
end
// Определяем матрицу ошибок R и заполняем ее нулями 
R(N+1,K+1)=0;
// Вычисляем коэффициент gamma
gam=a^2*delta/h^2;
r=1;
k=0;
// Цикл while для организации итерационного процесса при решении системы 
// уравнений методом Зейделя с точностью eps 
while r>eps
// Вычисление матрицы ошибок R во внутренних точках и пересчет значений 
// u во внутренних точках при решениии СЛАУ методом Зейделя
    for i=2:N
      for j=2:K+1
       R(i,j)=abs(u(i,j)-gam/(1+2*gam)*(u(i-1,j)+u(i+1,j))-u(i,j-1)/....
          (1+2*gam)-delta*f(x(i),t(j))/(1+2*gam));
       u(i,j)=gam/(1+2*gam)*(u(i-1,j)+u(i+1,j))+u(i,j-1)/....
          (1+2*gam)+delta*f(x(i),t(j))/(1+2*gam);
      end
    end
// Поиск максимума в матрице ошибок
    r=R(1,1);
    for i=1:N+1
        for j=1:K+1
            if R(i,j)>r 
                r=R(i,j);
            end
        end
    end
// Увеличение количества итерации
k=k+1;
end
disp(k)
endfunction

[U,X,T]=neyavn(50,200,5,3,0.4,0.1);
clf();
surf(X,T,U');
title('PARABOLIC EQUATION (var.2)');
xlabel('X');
ylabel('T');

halt;
clear;

// Использование неявной разностной схемы в случае, когда f(x,t)<>0, 
// рассмотрим на примере еще одной задачи.
// Найти распределение температуры в стержне длиной L, начальная 
// температура стержня задается произвольной функцией f(x). Температуры 
// концов стержня равны u(0,t)=U1=const, u(L,t)=U2=const.
// На боковой поверхности стержня происходит теплообмен по закону Ньютона
// со средой, температура которой равна u0.

//Начальное условие
function y=fi(x)
    y=exp(0.15*x)
endfunction

function [u,x,t,r,k]=neiav(N,K,L,T,a,h,U1,U2,u0,eps)
// Функция решения параболического уравнения методом сеток с помощью 
// неявной разностной схемы. 
// N - количество участков, на которые разбивается интервал по х (0,L); 
// K - количество участков, на которые разбивается интервал по t (0,T);
// a, h -  параметры диффренциального уравнения теплопроводности;
// eps - точность решения СЛАУ методом Зейделя;
// U1 - температура на левом конце стержня;
// U2 - температура на правом конце стержня;
// Функция neiav возвращает:
// u - матрицу решений в узлах сетки, массив x, массив t,
// r - точность решения системы методом Зейделя,
// к - количество итераций при решении системы методом Зейделя.

// Вычисляем шаг по х
hx=L/N;
// Вычисляем шаг по t
delta=T/K;
// Формируем массив x и первый столбец матрицы решений U из начального 
// условия
for i=1:N+1
    x(i)=(i-1)*hx;
    u(i,1)=fi(x(i));
end
// Формируем массив t, первую и последнюю строку матрицы решений U из 
// граничных условий
for j=1:K+1
    t(j)=(j-1)*delta;
    u(1,j)=U1;
    u(N+1,j)=U2;
end
// Определяем матрицу ошибок R и заполняем ее нулями 
R(N+1,K+1)=0;
// Вычисляем коэффициент gamma
gam=a^2*delta/hx^2;
r=1;
k=0;
// Цикл while для организации итерационного процесса при решении системы 
// уравнений методом Зейделя
// с точностью eps 
while r>eps
// Вычисление матрицы ошибок R во внутренних точках и пересчет значений 
// u во внутренних точках при решениии СЛАУ методом Зейделя
    for j=2:K+1
        for i=2:N
            V=gam*(u(i-1,j)+u(i+1,j))/(1+2*gam+delta*hx)+u(i,j-1)/....
                (1+2*gam+delta*hx)+delta*h*u0/(1+2*gam+delta*hx);
            R(i,j)=abs(V-u(i,j));
            u(i,j)=V;
        end
    end
// Поиск максимума в матрице ошибок
    r=R(1,1);
    for i=1:N+1
        for j=1:K+1
            if R(i,j)>r 
                r=R(i,j);
            end
        end
    end
// Увеличение количества итераций
    k=k+1;
end
endfunction

//Вызов функции решения задачи
[U,X,T,R,K]=neiav(50,200,5,3,0.4,0.5,1,2.117,30,0.001);
clf();
//Построение графика функции
surf(X,T,U');
title('Example 12.2');
xlabel('X');
ylabel('T');

halt;
clear;

// Использование метода сеток для решения гиперболических уравнений.
// Решим начально-граничную задачу.

//Правая часть дифференциального уравнения.
function y=f(x,t)
    y=sin(x*t)
endfunction

function y=fi(x)
    y=exp(0.15*x)
endfunction

function y=psi(x)
    y=exp(0.25*x)
endfunction

function [u,x,t]=ggg(N,K,L,T,a)
// Функция решения гиперболического уравнения с помощью явной разностной 
// схемы. 
// Входные данные:
// N - количество участков, на которые разбивается интервал по х (0,L); 
// K - количество участков, на которые разбивается интервал по t (0,T); 
// a - параметр дифференциального уравнения теплопроводности. 
// Выходные данные:
// u - матрица решений в узлах сетки, массив x, массив t.

// Вычисляем шаг по х
h=L/N;
// Вычисляем шаг по t
delta=T/K;
// Формируем массив x, первый и второй столбцы матрицы решений u из 
// начального условия
for i=1:N+1
    x(i)=(i-1)*h;
    u(i,1)=fi(x(i));
    u(i,2)=u(i,1)+delta*psi(x(i));
end
// Формируем массив t, первую и последнюю строку матрицы решений U из 
// граничных условий
for j=1:K+1
    t(j)=(j-1)*delta;
end
// Формируем первую и последнюю строку матрицы решений U из граничных 
// условий
for j=2:K+1
    u(1,j)=0;
    u(N+1,j)=fi(L);
end
gam=a^2*delta^2/h^2;
// Формируем матрицу решений u с помощью явной разностной схемы
for j=2:K
    for i=2:N
        u(i,j+1)=-u(i,j-1)+gam*u(i-1,j)+(2-2*gam)*u(i,j)+....
            gam*u(i+1,j)+delta^2*f(x(i),t(j));
    end
end
endfunction 
  
//Вызов функции решения задачи
[U,X,T]=ggg(50,200,5,3,0.4);
clf();
//Построение графика функции
surf(X,T,U');
title('Example ggg');
xlabel('X');
ylabel('T');

// Для решения гиперболических уравнений можно построить и неявные схемы, 
// однако в связи с тем, что у них нет таких преимуществ перед явными, 
// как в параболических уравнениях, их имеет смысл использовать только 
// когда нельзя построить явную, например, при смешанных условиях на 
// границе области

halt;
clear;

// Использование метода сеток для решения эллиптических уравнений.

function [Psi,x,y,k]=ellip(R,a,b,Nx,Ny,eps)
// Функция ellip решения задачи.
// Входные данные:
// R, a, b - значения, определяющие область решения задачи,
// Nx - количество участков, на которые разбивается интервал по х(R-b,R+b);
// Ny - количество участков, на которые разбивается интервал по y (-a,a);
// eps - точность решения уравнения (12.26) методом Зейделя.
// Выходные данные:
// Psi - матрица решений в узлах сетки, массив x, массив y,
// k - количество итерация при решении разностного уравнения методом Зейделя.

// Вычисляем шаг по y
hy=2*a/Ny;
// Вычисляем шаг по x
hx=2*b/Nx;
// Формируем массив x, первый и последний столбцы матрицы решений Psi из 
// граничного условия
for i=1:Nx+1
    x(i)=R-b+(i-1)*hx;
    Psi(i,1)=0;
    Psi(i,Ny+1)=0;
end;
// Формируем массив y, первую и последнюю строки матрицы решений Psi из 
// граничного условия
for j=1:Ny+1
    y(j)=-a+(j-1)*hy;
    Psi(1,j)=0;
    Psi(Nx+1,1)=0;
end;
// Вычисляем коэффициенты разностного уравнения
A=2/hy^2+2/hx^2;
D=1/hy^2;
for i=2:Nx+1
    B(i)=1/hx^2+5/(2*hx*x(i));
    C(i)=1/hx^2-5/(2*hx*x(i));
end
//Решение разностного уравнения методом Зейделя с точностью eps
p=1;
k=0;
while p>eps
    for i=2:Nx
        for j=2:Ny
            V=1/A*(B(i)*Psi(i-1,j)+C(i)*Psi(i+1,j)+D*(Psi(i,j-1)+...
                Psi(i,j+1))+2);
            R(i,j)=abs(V-Psi(i,j));
            Psi(i,j)=V;
        end
    end 
    p=R(2,2);
    for i=2:Nx
        for j=2:Ny
            if R(i,j)>p
                p=R(i,j);
            end
        end
    end
    k=k+1;
end
endfunction

//Вызов функции решения задачи
[PSI,X,Y,K]=ellip(18,3,6,32,16,0.01);
//Построение графика функции
clf();
surf(X,Y,PSI');
title('Example el');
xlabel('X');
ylabel('Y');

// В случае медленной сходимости итерационных процессов при решении сеточных 
// уравнений, получаемых при аппроксимации, имеет смысл попробовать заменить 
// метод Зейделя градиентными методами или методами релаксации.  

// Метод сеток позволяет решить широкий класс уравнений в частных производных. 
// Однако при сложной геометрии области, уравнениях с переменными коэффициентами, 
// сложными условиями на границе области использование этого метода не 
// целесообразно. В этих случаях можно использовать метод конечных элементов, 
// который реализован в кроссплатформенном свободно распространяемом пакете 
// freefem.


clear
